
// A PETSCII drawing app. Originally made for Zoo'13 PETSCII compo, 'coz I couln't find anything
// suitable for Linux/Mac. -Marq

// 28.9.2013: Self-contained C export, BASIC export
// 28.9.2013: Big cleanup, comments changed to English, improved paste
// 27.9.2013: Border scales too, show target of paste, preview mode
// 27.9.2013: Save, automatic backup, load, initial copy paste
// 26.9.2013: The first version

/* "Help"
 
    lmb    - draw, select colors and chars
    mmb    - eraser
    rmb    - pick char+color
    
    s      - invert char
    x      - cut to clipboard
    c      - paste at cursor
    space  - unselect
    tab    - preview
    shift  - coloring mode, just change the color
    ctrl   - selection mode
    
    w      - save file (just data)
    W      - export file as self-contained C source
    b      - export file as self-contained BASIC source

*/

// Basic settings
final int zoom=2,     // Changes everything in the UI
          X=40,Y=25,  // Picture size in chars
          CHAR=8,     // Character size in pixels
          
          borderpos=CHAR,// Locations of the selectors
          bgpos=3*CHAR,
          penpos=6*CHAR,
          charselpos=9*CHAR,
          
          backup=60*60*2; // How many frames until automatic backup (2 min)
          
final String FILENAME="img.c";

int chars[],colors[],
    clip_chars[],clip_colors[];

int border=14,bg=6,pen=1, // Drawing colors
    current=0,
    backupcounter=backup,
    borderw,
    selx=0,sely=0,selw=0,selh=0; // Selection params

boolean shifted=false,
        controlled=false,
        firstsel=true,
        preview=false;

int rgb[]={#000000,#FFFFFF,#68372B,#70A4B2,#6F3D86,#588D43,#352879,#B8C76F,#6F4F25,#433900,
           #9A6759,#444444,#6C6C6C,#9AD284,#6C5EB5,#959595};

PImage petscii,pet[];
PFont font;

void setup()
{   
    borderw=25*zoom;
    
    size(X*CHAR*zoom+(X-1)+borderw*3+16*(1+CHAR*zoom), Y*CHAR*zoom+(Y-1)+borderw*2);
    noStroke();
    noSmooth();
        
    font=loadFont("arial.vlw"); // 16 pix
    textFont(font);
    
    chars=new int[X*Y];
    colors=new int[X*Y];
    clip_chars=new int[X*Y];
    clip_colors=new int[X*Y];
    
    for(int i=0;i<X*Y;i++)
    {
        chars[i]=32;
        colors[i]=14;
    }
    
    // Generate chars from the image
    petscii=loadImage("petsciiwide.png");
    pet=new PImage[128];
    petscii.loadPixels();
    
    for(int i=0;i<128;i++) // The second half is drawn in inverse
    {
        pet[i]=new PImage(CHAR,CHAR,ARGB);
        pet[i].loadPixels();
        
        for(int y=0;y<CHAR;y++)
            for(int x=0;x<CHAR;x++)
            {
                if((petscii.pixels[i*CHAR+x+y*CHAR*128]&255)<20) // Black
                    pet[i].pixels[y*CHAR+x]=0x00000000;
                else
                    pet[i].pixels[y*CHAR+x]=0xffffffff;
            }
            
        pet[i].updatePixels();
    }
    
    // Load previous image (comment out to create a new one)
    load_c(FILENAME);
}

void draw()
{
    int uix=X*CHAR*zoom+(X-1)+borderw*2,  // UI elements x location
        blox=(mouseX-borderw)/(zoom*CHAR+1), // Mouse coordinates in character blocks
        bloy=(mouseY-borderw)/(zoom*CHAR+1),
        selectx=0,selecty=0;
     
    boolean infield=false;
    
    // In the drawing area?
    if(blox>=0 && bloy>=0 && blox<X && bloy<Y)
        infield=true;
    
    // Border
    background(rgb[border]);
    // Grid bg
    fill(#555555);
    rect(borderw,borderw,uix-borderw*2,height-borderw*2);
    
    // Draw the chars
    for(int y=0,i=0;y<Y;y++)
        for(int x=0;x<X;x++,i++)
            drawchar(x*(zoom*CHAR+1)+borderw,y*(zoom*CHAR+1)+borderw,chars[i],colors[i],bg,false);
    
    // Draw char selector
    fill(#555555);
    rect(uix,borderw+charselpos*zoom,16*(1+CHAR*zoom),16*(1+CHAR*zoom));
    for(int y=0,i=0;y<16;y++)
        for(int x=0;x<16;x++,i++)
        {
            drawchar(uix+x*(1+CHAR*zoom),borderw+charselpos*zoom+y*(1+CHAR*zoom),i,pen,bg,false);
            if(i==current)
            {
                selectx=uix+x*(1+CHAR*zoom);
                selecty=borderw+charselpos*zoom+y*(1+CHAR*zoom);
            }
        }
    
    // Selected char
    stroke(#ff0000);
    noFill();
    rect(selectx,selecty,CHAR*zoom+1,CHAR*zoom+1);
    noStroke();
    
    // Color selectors
    for(int x=0;x<16;x++)
    {
        fill(rgb[x]);
        rect(uix+x*CHAR*zoom,borderw+borderpos*zoom, CHAR*zoom,CHAR*zoom);
        rect(uix+x*CHAR*zoom,borderw+bgpos*zoom,     CHAR*zoom,CHAR*zoom);
        rect(uix+x*CHAR*zoom,borderw+penpos*zoom,    CHAR*zoom,CHAR*zoom);
    }
    
    stroke(128);
    noFill();
    rect(uix,borderw+borderpos*zoom, 16*CHAR*zoom,CHAR*zoom);
    rect(uix,borderw+bgpos*zoom,     16*CHAR*zoom,CHAR*zoom);
    rect(uix,borderw+penpos*zoom,    16*CHAR*zoom,CHAR*zoom);
    
    noStroke();
    fill(rgb[border]^0x808080);
    text("border",     uix,borderw+borderpos*zoom);
    text("background", uix,borderw+bgpos*zoom);
    text("pen",        uix,borderw+penpos*zoom);

    // Active color markers    
    noFill();
    stroke(#ff0000);
    rect(uix+CHAR*zoom*border, borderw+borderpos*zoom, zoom*CHAR,zoom*CHAR);
    rect(uix+CHAR*zoom*bg,     borderw+bgpos*zoom,     zoom*CHAR,zoom*CHAR);
    rect(uix+CHAR*zoom*pen,    borderw+penpos*zoom,    zoom*CHAR,zoom*CHAR);
    noStroke();

    boolean erasing=false;
    
    // User interaction(!)
    if(mousePressed && !preview)
    {
        // Color selectors
        if(mouseX>=uix && mouseX<uix+16*CHAR*zoom)
        {
            if(mouseY>=borderw+borderpos*zoom && mouseY<borderw+borderpos*zoom+CHAR*zoom)
                border=(mouseX-uix)/(CHAR*zoom);
            if(mouseY>=borderw+bgpos*zoom && mouseY<borderw+bgpos*zoom+CHAR*zoom)
                bg=(mouseX-uix)/(CHAR*zoom);
            if(mouseY>=borderw+penpos*zoom && mouseY<borderw+penpos*zoom+CHAR*zoom)
                pen=(mouseX-uix)/(CHAR*zoom);
        }
        
        // Inside drawing area
        if(infield)
        {
            if(controlled) // Selection going on
            {
                if(firstsel)
                {
                    firstsel=false;
                    selx=blox;
                    sely=bloy;
                }
                selw=blox-selx+1;
                selh=bloy-sely+1;
                if(selw<0)
                    selw=0;
                if(selh<0)
                    selh=0;
                 
                // Copy automatically   
                int k=0;
                for(int i=0;i<selh;i++)
                    for(int j=0;j<selw;j++,k++)
                    {
                        clip_chars[k]=chars[(i+sely)*X+(j+selx)];
                        clip_colors[k]=colors[(i+sely)*X+(j+selx)];
                    }
            }
            else // Normal operation
            {
                int idx=(mouseX-borderw)/(zoom*CHAR+1)+(mouseY-borderw)/(zoom*CHAR+1)*X;
                
                if(mouseButton==LEFT)
                {
                    if(!shifted)
                        chars[idx]=current;
                    colors[idx]=pen;
                }
                if(mouseButton==RIGHT)
                {
                    current=chars[idx];
                    pen=colors[idx];
                }
                if(mouseButton==CENTER)
                {
                    chars[idx]=32;
                    erasing=true;
                }
            }
        }
        
        // Char selector
        if(mouseX>=uix && mouseY>=borderw+charselpos*zoom && mouseX<uix+16*(1+CHAR*zoom) &&
           mouseY<borderw+charselpos*zoom+16*(1+CHAR*zoom))
        {
            current=(mouseX-uix)/(zoom*CHAR+1)+(mouseY-(borderw+charselpos*zoom))/(zoom*CHAR+1)*16;
        }
    }
    else
        firstsel=true;
    
    // Show the upcoming character
    if(!controlled && infield)
    {
        if(shifted)
        {
            int idx=(mouseX-borderw)/(zoom*CHAR+1)+(mouseY-borderw)/(zoom*CHAR+1)*X;
            
            drawchar((mouseX-borderw)/(CHAR*zoom+1)*(CHAR*zoom+1)+borderw,
                     (mouseY-borderw)/(CHAR*zoom+1)*(CHAR*zoom+1)+borderw,chars[idx],pen,bg,false);
        }
        else
        {
            int tmp=current;
            if(erasing)
                tmp=32;
            
            drawchar((mouseX-borderw)/(CHAR*zoom+1)*(CHAR*zoom+1)+borderw,
                     (mouseY-borderw)/(CHAR*zoom+1)*(CHAR*zoom+1)+borderw,tmp,pen,bg,false);
        }
    }
    
    // Show selection
    if(selw!=0 && selh!=0)
    {
        noFill();
        stroke(#ff0000);
        rect(borderw+selx*(1+CHAR*zoom)-1,borderw+sely*(1+CHAR*zoom)-1,
             selw*(1+CHAR*zoom),selh*(1+CHAR*zoom));

        // Show paste target
        stroke(0,255,0,100);
        if(firstsel && infield)
        {
            int halfx=blox-selw/2,
                halfy=bloy-selh/2;
            
            int left=max(borderw+halfx*(1+CHAR*zoom)-1,borderw),
                top=max(borderw+halfy*(1+CHAR*zoom)-1,borderw),
                right=min(borderw+(halfx+selw)*(1+CHAR*zoom),uix-borderw),
                bottom=min(borderw+(halfy+selh)*(1+CHAR*zoom),height-borderw);
                
            rect(left,top,right-left,bottom-top);
        }
        
        noStroke();
    }
    
    // Preview
    if(preview)
    {
        // Borderi
        int xpos=width/2-X*zoom*CHAR/2,
            ypos=height/2-Y*zoom*CHAR*5/4/2;
        
        background(rgb[border]);
        
        for(int y=0,i=0;y<Y;y++)
            for(int x=0;x<X;x++,i++)
                drawchar(x*(zoom*CHAR)+xpos,y*(zoom*CHAR*5/4)+ypos,chars[i],colors[i],bg,true);
    }
    
    // Backup?
    backupcounter--;
    if(backupcounter==0)
    {
        backupcounter=backup;
        save_c("_backup_.c",false);
    }
    
    // Better do this at times or we might run out of memory
    if(frameCount%100==0)
        System.gc();
}

void keyPressed() // Keyboard commands
{
    if(keyCode==SHIFT)
        shifted=true;

    if(keyCode==CONTROL)
        controlled=true;
        
    if(key=='s') // Invert
        current^=128;
        
    if(key=='w')
        save_c(FILENAME,false);
    if(key=='W')
        save_c(EXPORT_C,true);
    if(key=='b')
        save_bas(EXPORT_BAS);
        
    if(key==TAB)
        preview=!preview;

    if(key=='x') // Cut
    {
        int k=0;
        for(int i=0;i<selh;i++)
            for(int j=0;j<selw;j++,k++)
            {
                clip_chars[k]=chars[(i+sely)*X+(j+selx)];
                clip_colors[k]=colors[(i+sely)*X+(j+selx)];
                
                chars[(i+sely)*X+(j+selx)]=32;
                colors[(i+sely)*X+(j+selx)]=pen;
            }
    }

    if(key=='c') // Paste from clipboard
    {
        int blox=(mouseX-borderw)/(zoom*CHAR+1),
            bloy=(mouseY-borderw)/(zoom*CHAR+1);
            
        if(blox>=0 && bloy>=0 && blox<X && bloy<Y)
        {
            for(int i=0;i<selh;i++)
            {
                for(int j=0;j<selw;j++)
                {
                    int x=blox-selw/2+j,
                        y=bloy-selh/2+i;
                        
                    if(x>=0 && y>=0 && x<X & y<Y)
                    {
                        chars[x+y*X]=clip_chars[j+i*selw];
                        colors[x+y*X]=clip_colors[j+i*selw];
                    }
                }
            }
        }
    }

    if(key==' ') // Unselect
        selw=selh=0;

    // Better disable this to prevent accidents
    if(key==ESC)
        key=0;
}

void keyReleased()
{
    if(keyCode==SHIFT)
        shifted=false;
        
    if(keyCode==CONTROL)
        controlled=false;
}

// Plot a char here with this color
void drawchar(int x,int y,int num,int fg,int bg,boolean stretch)
{
    if(num>=128) // Inverted PETSCII chars
    {
        int tmp=fg;
        fg=bg;
        bg=tmp;
        num-=128;
    }
    
    fill(rgb[bg]); // Background
    if(stretch)
        rect(x,y,zoom*CHAR,zoom*CHAR*5/4);
    else
        rect(x,y,zoom*CHAR,zoom*CHAR);
 
    tint(rgb[fg]); // Character pixels
    if(stretch)
        image(pet[num],x,y,CHAR*zoom,CHAR*5/4*zoom);
    else
        image(pet[num],x,y,CHAR*zoom,CHAR*zoom);
}

// Save a piccy
final String EXPORT_C="export.c";

void save_c(String name,boolean selfcontained)
{
    PrintWriter f=createWriter(name);
    
    if(selfcontained)
        f.println(C_HEADER);
   
    f.println("unsigned char img[]={ // border,bg,chars,colors");
    f.println(str(border)+","+str(bg)+",");
    
    for(int y=0;y<Y;y++)
    {
        for(int x=0;x<X;x++)
            f.print(str(chars[y*X+x])+",");
        f.println();
    }
    for(int y=0;y<Y;y++)
    {
        for(int x=0;x<X;x++)
        {
            f.print(str(colors[y*X+x]));
            if(y!=Y-1 || x!=X-1)
                f.print(",");
        }
        f.println();
    }
    f.println("};");
    
    if(selfcontained)
        f.println(C_FOOTER);
    
    f.flush();
    f.close();
    
    println("Written "+name);
}

// Load a piccy
void load_c(String name)
{
    String lines[]=loadStrings(name),
           s[]=splitTokens(lines[1],",");    
    int i=2;

    border=int(s[0]);
    bg=int(s[1]);
    
    for(int y=0,j=0;y<Y;y++,i++)
    {
        s=splitTokens(lines[i],",");
        for(int x=0;x<X;x++,j++)
            chars[j]=int(s[x]);
    }
    for(int y=0,j=0;y<Y;y++,i++)
    {
        s=splitTokens(lines[i],",");
        for(int x=0;x<X;x++,j++)
            colors[j]=int(s[x]);
    }
    
    println("Loaded "+name);
}

final String C_HEADER=
"// Compile with cc65: cl65 export.c -o export.prg\n"+
"\n"+
"#include <string.h>\n";

final String C_FOOTER=
"\n"+
"void main(void)\n"+
"{\n"+
"  *(char *)0xd011=8+3;\n"+
"  *(char *)0xd020=img[0];\n"+
"  *(char *)0xd021=img[1];\n"+
"  *(char *)0xd018=0x14;\n"+
"\n"+
"  memcpy((void *)0x400,&img[2],1000);\n"+
"  memcpy((void *)0xd800,&img[2+1000],1000);\n"+
"\n"+
"  *(char *)0xd011=16+8+3;\n"+
"\n"+
"  while(1);\n"+
"}";

// Save as plaintext self-contained BASIC file
final String EXPORT_BAS="export.bas";

void save_bas(String name)
{
    PrintWriter f=createWriter(name);
    
    f.println("10 rem petcat -text -w2 -o export.prg export.bas");
    f.println("20 print chr$(147)");
    f.println("30 poke 53280,"+str(border));
    f.println("40 poke 53281,"+str(bg));
    f.println("50 for i=0 to 999:read a:poke 1024+i,a:read a:poke 55296+i,a:next");
    f.println("60 goto 60");
    
    int line=70,idx=0;
    for(int i=0;i<chars.length;i+=8)
    {
        f.print(str(line)+" data ");
        for(int j=0;j<8;j++,idx++)
        {
            f.print(str(chars[idx])+",");
            f.print(colors[idx]);
            if(j!=7)
                f.print(",");
        }
        f.println();
        
        line+=10;
    }
    
    f.println(str(line)+" end");

    f.flush();
    f.close();
    
    println("Written "+name);
}
